/*
        wavfir.c
        FIR-Filter auf WAV-Datei anwenden

        Stand: 25.10.2021
        max. 8192 FIR-Koeffizienten
*/

# define MAX_FIR_LENGTH 8192        /* max. Anzahl von FIR-Koeffizienten */

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>


static double fir_coef1[MAX_FIR_LENGTH], fir_coef2[MAX_FIR_LENGTH],
              data1[MAX_FIR_LENGTH], data2[MAX_FIR_LENGTH];


unsigned long char2long(unsigned char *buffer)
 {
  int i;
  unsigned long value;

  value = 0;
  for (i=0; i<4; i++)
    value |= buffer[i] << (i*8);
  return(value);
 }


int read_bytes(FILE *stream, char *buffer, unsigned long n)
 {
  int c;
  unsigned long i;

  for (i=0; i<n; i++)
    if ((c = fgetc(stream)) == EOF)
      return(-1);
    else
      buffer[i] = c;
  return(0);
 }


unsigned char long2char(long number, int index)
 {
  return((number >> (index*8)) & 0xff);
 }


void err_exit(char message[], char msg_arg[], int err_code)
 {
  if (msg_arg == NULL)
    fprintf(stderr, message);
  else
    fprintf(stderr, message, msg_arg);
  exit(err_code);
 }


        /* return code: 0 = OK, 1 = Dateiende */
int read_sample(FILE *stream, int bits_per_sample, double *value)
 {
  short c1, c2;

  if (bits_per_sample == 8)
   {
    if ((c1 = fgetc(stream)) == EOF)
      return(1);
    *value = 2.0*c1/255.0 - 1.0;
   }
  else
   {
    if ((c1 = fgetc(stream)) == EOF)
      return(1);
    if ((c2 = fgetc(stream)) == EOF)
      return(1);
    c1 = c1 | (c2 << 8);
    *value = (2.0*c1+1.0)/65535.0;
   }
  return(0);
 }


int write_sample(double value, int bits_per_sample, FILE *stream)
 {
  int sample, clipped;

  clipped = 0;
  if (value > 1.0)
   {
    value = 1.0;
    clipped = 1;
   }
  if (value < -1.0)
   {
    value = -1.0;
    clipped = 1;
   }

  if (bits_per_sample == 8)
   {
    sample = rint((value+1.0)*255/2.0);
    if (fprintf(stream, "%c", sample) <= 0)
      err_exit("wavfir: Can't write to output file. File system is possibly full.\n", NULL, 2);
   }
  else
   {
    sample = rint((value*65535.0-1.0)/2.0);
    if (fprintf(stream, "%c", sample & 0xff) <= 0)
      err_exit("wavfir: Can't write to output file. File system is possibly full.\n", NULL, 2);
    if (fprintf(stream, "%c", (sample >> 8) & 0xff) <= 0)
      err_exit("wavfir: Can't write to output file. File system is possibly full.\n", NULL, 2);
   }

  return(clipped);
 }


long fir(FILE *in_stream, FILE *out_stream, int channels, int i_bits,
         int o_bits, int fir_samples, int force_stereo, long in_samples)
 {
  register int i, pos2;
  int pos, k, l, no_more_data, clipped, m;
  register double sum, sum2;
  double max;
  long out_samples;

  clipped = out_samples = 0;
  max = 0.0;
  if (channels == 1)        /* Mono */
   {
    k = pos = 0;                /* Teil1: Einschwingen */
    do
     {
      if ((no_more_data = read_sample(in_stream, i_bits, &(data1[pos]))) == 0)
       {
        sum = 0.0;
        for (i=0; i<=pos; i++)
          sum += fir_coef1[i] * data1[pos-i];
        if (fabs(sum) > max)
          max = fabs(sum);
        clipped |= write_sample(sum, o_bits, out_stream);
        if (force_stereo)
         {
          sum = 0.0;
          for (i=0; i<=pos; i++)
            sum += fir_coef2[i] * data1[pos-i];
          if (fabs(sum) > max)
            max = fabs(sum);
          clipped |= write_sample(sum, o_bits, out_stream);
         }
        out_samples++;
        pos++;
        k++;
        if (out_samples == in_samples)
          no_more_data = 1;
       }
     }
    while (!no_more_data && (pos < fir_samples));
    if (!no_more_data)                /* Teil2: Mittelteil */
     {
      pos = 0;
      do
       {
        if ((no_more_data = read_sample(in_stream, i_bits, &(data1[pos]))) == 0)
         {
          sum = 0.0;
          pos2 = pos;
          for (i=0; i<fir_samples; i++)
           {
            sum += fir_coef1[i] * data1[pos2];
            if (--pos2 < 0)
              pos2 = fir_samples-1;
           }
          if (fabs(sum) > max)
            max = fabs(sum);
          clipped |= write_sample(sum, o_bits, out_stream);
          if (force_stereo)
           {
            sum = 0.0;
            pos2 = pos;
            for (i=0; i<fir_samples; i++)
             {
              sum += fir_coef2[i] * data1[pos2];
              if (--pos2 < 0)
                pos2 = fir_samples-1;
             }
            if (fabs(sum) > max)
              max = fabs(sum);
            clipped |= write_sample(sum, o_bits, out_stream);
           }
          out_samples++;
          if (++pos == fir_samples)
            pos = 0;
          if (out_samples == in_samples)
            no_more_data = 1;
          if ((in_samples > 100) && (out_samples % (in_samples/100) == 0))
            fprintf(stderr, "\r%3ld%% done", out_samples*100/in_samples);
         }
       }
      while (!no_more_data);
     }
    if (--pos < 0)                /* letzter Sample konnte nicht belegt werden */
      pos = fir_samples-1;
    m = fir_samples-1;        // noch zu verarbeitender Teil der Impulsantwort
    for (l=1; l<fir_samples; l++)        /* Teil3: Ausklingen */
     {
      sum = 0.0;
      pos2 = pos;
      for (i=0; i<m; i++)
       {
        if (i < k)        // wenn noch ausreichend Daten fuer Impulsantwort
         {
          sum += fir_coef1[l+i] * data1[pos2];
          if (--pos2 < 0)
            pos2 = fir_samples-1;
         }
       }
      if (fabs(sum) > max)
        max = fabs(sum);
      clipped |= write_sample(sum, o_bits, out_stream);
      if (force_stereo)
       {
        sum = 0.0;
        pos2 = pos;
        for (i=0; i<m; i++)
         {
          if (i < k)        // wenn noch ausreichend Daten fuer Impulsantwort
           {
            sum += fir_coef2[l+i] * data1[pos2];
            if (--pos2 < 0)
              pos2 = fir_samples-1;
           }
         }
        if (fabs(sum) > max)
          max = fabs(sum);
        clipped |= write_sample(sum, o_bits, out_stream);
       }
      out_samples++;
      m--;
     }
   }
  else                /* Stereo */
   {
    k = pos = 0;                /* Teil1: Einschwingen */
    do
     {
      if ((no_more_data = read_sample(in_stream, i_bits, &(data1[pos]))
           | read_sample(in_stream, i_bits, &(data2[pos]))) == 0)
       {
        sum = sum2 = 0.0;
        for (i=0; i<=pos; i++)
         {
          sum += fir_coef1[i] * data1[pos-i];
          sum2 += fir_coef2[i] * data2[pos-i];
         }
        if (fabs(sum) > max)
          max = fabs(sum);
        if (fabs(sum2) > max)
          max = fabs(sum2);
        clipped |= write_sample(sum, o_bits, out_stream);
        clipped |= write_sample(sum2, o_bits, out_stream);
        out_samples++;
        pos++;
        k++;
       }
     }
    while (!no_more_data && (pos < fir_samples));
    if (!no_more_data)                /* Teil2: Mittelteil */
     {
      pos = 0;
      do
       {
        if ((no_more_data = read_sample(in_stream, i_bits, &(data1[pos]))
             | read_sample(in_stream, i_bits, &(data2[pos]))) == 0)
         {
          sum = sum2 = 0.0;
          pos2 = pos;
          for (i=0; i<fir_samples; i++)
           {
            sum += fir_coef1[i] * data1[pos2];
            sum2 += fir_coef2[i] * data2[pos2];
            if (--pos2 < 0)
              pos2 = fir_samples-1;
           }
          if (fabs(sum) > max)
            max = fabs(sum);
          if (fabs(sum2) > max)
            max = fabs(sum2);
          clipped |= write_sample(sum, o_bits, out_stream);
          clipped |= write_sample(sum2, o_bits, out_stream);
          out_samples++;
          if ((in_samples > 100) && (out_samples % (in_samples/100) == 0))
            fprintf(stderr, "\r%3ld%% done", out_samples*100/in_samples);
          if (++pos == fir_samples)
            pos = 0;
         }
       }
      while (!no_more_data);
     }
    if (--pos < 0)                /* letzter Sample konnte nicht belegt werden */
      pos = fir_samples-1;
    m = fir_samples-1;        // noch zu verarbeitender Teil der Impulsantwort
    for (l=1; l<fir_samples; l++)        /* Teil3: Ausklingen */
     {
      sum = sum2 = 0.0;
      pos2 = pos;
      for (i=0; i<m; i++)
       {
        if (i >= k)
          break;
        sum += fir_coef1[l+i] * data1[pos2];
        sum2 += fir_coef2[l+i] * data2[pos2];
        if (--pos2 < 0)
          pos2 = fir_samples-1;
       }
      if (fabs(sum) > max)
        max = fabs(sum);
      if (fabs(sum2) > max)
        max = fabs(sum2);
      clipped |= write_sample(sum, o_bits, out_stream);
      clipped |= write_sample(sum2, o_bits, out_stream);
      out_samples++;
      m--;
     }
   }

  fprintf(stderr, "\rPeak level was %.0f%%.\n", max*100);
  if (clipped)
    fprintf(stderr, "\n*** wavfir: Warning: output file was clipped.\n");
  return(out_samples);
 }


/*----------------------------------------------------------------------*/

int main(int argc, char *argv[])
 {
  int k, force_stereo;
  unsigned int i, i_bits, o_bits, n, channels, o_channels;
  long length, freq, samples, l;
  double x, t;
  FILE *in_stream, *out_stream;
  static char line[80];

  if ((argc == 2) && (strcmp(argv[1], "-h") == 0))
   {
    printf("wavfir - apply an FIR filter on a WAV file.\n");
    printf("Usage: wavfir [-8 | -16] fir-file [input-file [output-file]]\n");
    return(0);
   }

  if (argc < 2)
    err_exit("wavfir: Bad arguments. Type 'wavfir -h' for help.\n", NULL, 1);

  if (strcmp(argv[1], "-8") == 0)
   {
    o_bits = 8;
    i = 1;
   }
  else if (strcmp(argv[1], "-16") == 0)
   {
    o_bits = 16;
    i = 1;
   }
  else
    o_bits = i = 0;

  if ((argc < 2+i) || (argc > 4+i))
    err_exit("wavfir: Bad arguments. Type 'wavfir -h' for help.\n", NULL, 1);


        /*- - - - - - - - - - FIR-Datei laden - - - - - - - - - - - - -*/

  if ((in_stream=fopen(argv[1+i], "r")) == NULL)
    err_exit("wavfir: Can't open FIR file '%s'.\n", argv[1+i], 2);

  do                        /* Kommentar-Kopf ueberspringen */
    if (fgets(line, 80, in_stream) == NULL)
      err_exit("wavfir: Bad FIR file.\n", NULL, 2);
  while (sscanf(line, "%lf", &x) != 1);

  n = channels = 0;
  do
   {
    k = sscanf(line, "%lf %lf", &(fir_coef1[n]), &(fir_coef2[n]));
    if (k > 0)
     {
      if (channels == 0)
        channels = k;
      else if (channels != k)
        err_exit("wavfir: Number of columns in FIR file is not constant.\n",
                 NULL, 2);
      if (channels == 1)
        fir_coef2[n] = fir_coef1[n];
      n++;
     }
   }
  while ((k > 0) && (fgets(line, 80, in_stream) != NULL) && (n < MAX_FIR_LENGTH));

  if ((n == MAX_FIR_LENGTH) && (fgets(line, 80, in_stream) != NULL)
      && (sscanf(line, "%lf", &x) == 1))
    fprintf(stderr, "wavfir: FIR file truncated (max. %d samples).\n",
            MAX_FIR_LENGTH);

  fclose(in_stream);
  if (channels == 1)
   {
    force_stereo = 0;
    fprintf(stderr, "%d FIR samples, mono\n", n);
   }
  else
   {
    force_stereo = 1;
    fprintf(stderr, "%d FIR samples, stereo\n", n);
   }

        /*- - - - - - - Ein- und Ausgabedatei oeffnen - - - - - - - - - -*/

  if (argc < 3+i)
    in_stream = stdin;
  else
    if ((in_stream=fopen(argv[2+i], "rb")) == NULL)
      err_exit("wavfir: Can't open input file '%s'.\n", argv[2+i], 2);

  if (argc < 4+i)
    out_stream = stdout;
  else
    if ((out_stream=fopen(argv[3+i], "wb")) == NULL)
      err_exit("wavfir: Can't open output file '%s'.\n", argv[3+i], 2);


        /*- - - - - - - - - - WAV-Header lesen - - - - - - - - - - - -*/

  if (read_bytes(in_stream, line, 8) || (strncmp(line, "RIFF", 4) != 0))
    err_exit("wavfir: Not a WAV file.\n", NULL, 2);

  length = char2long(&(line[4]));
  if (length < 36)
    fprintf(stderr, "wavfir: Warning: Bad length in WAV header.\n");

  if (read_bytes(in_stream, line, 8) || (strncmp(line, "WAVEfmt ", 8) != 0))
    err_exit("wavfir: Corrupted WAV header.\n", NULL, 2);

  if (read_bytes(in_stream, line, 8))
    err_exit("wavfir: Corrupted WAV header.\n", NULL, 2);
  length = char2long(line);
  channels = line[6];
  if ((channels < 1) || (channels > 2))
    err_exit("wavfir: Number of channels (%d) not supported.\n", (char *)channels, 2);

  if (read_bytes(in_stream, line, 12))
    err_exit("wavfir: Corrupted WAV header.\n", NULL, 2);
  freq = char2long(line);
  i_bits = line[10];
  if ((i_bits != 8) && (i_bits != 16))
    err_exit("wav2dat: Unsupported bits/sample (%d) in WAV file.\n", (char *)i_bits, 2);
  if (o_bits == 0)
    o_bits = i_bits;

  if (length > 16)        /* ggf. Rest vom Chunk 'WAVEfmt ' ueberlesen */
    if (fseek(in_stream, length-16, SEEK_CUR) != 0)
     {
      fprintf(stderr, "wav2dat: Corrupted header.\n");
      exit(1);
     }

  do
   {
    if (read_bytes(in_stream, line, 8))
      err_exit("wavfir: Corrupted WAV header.\n", NULL, 2);
    length = char2long(&(line[4]));
    if (strncmp(line, "data", 4) != 0)
      if (fseek(in_stream, length, SEEK_CUR) != 0)
        err_exit("wavfir: Corrupted WAV header.\n", NULL, 2);
   }
  while (strncmp(line, "data", 4) != 0);

  samples = length / channels * 8 / i_bits;

  if (channels == 1)
    fprintf(stderr, "Input file: %ld samples, %ld Hz, mono, %d bits/sample.\n",
            samples, freq, i_bits);
  else
    fprintf(stderr, "Input file: %ld samples, %ld Hz, stereo, %d bits/sample.\n",
            samples, freq, i_bits);


        /*- - - - - - - - - - WAV-Header schreiben - - - - - - - - - - -*/

  if (force_stereo)
    o_channels = 2;
  else
    o_channels = channels;

  length = (samples+n-1)*(o_bits/8)*o_channels+36;
  fprintf(out_stream, "RIFF%c%c%c%c", long2char(length, 0), long2char(length, 1),
         long2char(length, 2), long2char(length, 3));

  fprintf(out_stream, "WAVEfmt %c%c%c%c%c%c", 16,0,0,0,1,0);

  fprintf(out_stream, "%c%c", o_channels, 0);

  fprintf(out_stream, "%c%c%c%c", long2char(freq, 0), long2char(freq, 1),
         long2char(freq, 2), long2char(freq, 3));

  l = freq*o_channels*(o_bits/8);
  fprintf(out_stream, "%c%c%c%c", long2char(l, 0), long2char(l, 1),
         long2char(l, 2), long2char(l, 3));

  fprintf(out_stream, "%c%c", (o_bits/8)*o_channels, 0);

  fprintf(out_stream, "%c%c", o_bits, 0);

  length -= 36;
  if (fprintf(out_stream, "data%c%c%c%c", long2char(length, 0), long2char(length, 1),
              long2char(length, 2), long2char(length, 3)) == 0)
    err_exit("wavfir: Can't write to file. (File system full?)\n", NULL, 2);

  if (o_channels == 1)
    fprintf(stderr, "Output file: %ld samples, %ld Hz, mono, %d bits/sample.\n",
            samples+n-1, freq, o_bits);
  else
    fprintf(stderr, "Output file: %ld samples, %ld Hz, stereo, %d bits/sample.\n",
            samples+n-1, freq, o_bits);

  fprintf(stderr, "%ld samples written.\n",
          fir(in_stream, out_stream, channels, i_bits, o_bits, n, force_stereo, samples));

  return(0);
 }

